#include <chrono>
#include <cmath>
#include <cstdint>
#include <mavsdk/mavsdk.h>
#include <mavsdk/plugins/action/action.h>
#include <mavsdk/plugins/mavlink_passthrough/mavlink_passthrough.h>
#include <mavsdk/plugins/telemetry/telemetry.h>
#include <iostream>
#include <fstream>
#include <functional>
#include <future>
#include <memory>
#include <thread>
#include "autopilot_wrapper.h"

using namespace mavsdk;
using std::chrono::seconds;
using std::this_thread::sleep_for;

#define ERROR_CONSOLE_TEXT "\033[31m" // Turn text on console red
#define NORMAL_CONSOLE_TEXT "\033[0m" // Restore normal console colour

struct Coordinates {
  double latitude;
  double longitude;
};

static std::ofstream log_file_fd;

static Mavsdk _mavsdk;
static Telemetry * telemetry;
static Action * action;
static MavlinkPassthrough * mavlink_passthrough;
static std::shared_ptr<System> msystem;

static auto prom = std::promise<std::shared_ptr<System>>{};
static std::future<std::shared_ptr<System>> fut;

static const float DEFAULT_RADIUS = 100;
static const float EARTH_RADIUS = 6371000;
static const float ADDED_DISTANCE = (2 * DEFAULT_RADIUS) / EARTH_RADIUS;
static const double COS_ADDED_DISTANCE = cos(ADDED_DISTANCE);
static const double SIN_ADDED_DISTANCE = sin(ADDED_DISTANCE);

static int mavsdk_started = 0;
static Telemetry::Position origin;

static Coordinates targeted_destination;
static Coordinates projected_destination;

// Logs functions

static void log(std::string message) {
    log_file_fd << message << std::endl;
}

static void log_error(std::string message) {
    log(ERROR_CONSOLE_TEXT + message + NORMAL_CONSOLE_TEXT);
}

template <class Enumeration>
static void log_error_from_result(std::string message, Enumeration result) {
    std::ostringstream oss;
    oss << message << ": " << result;
    log_error(oss.str());
}

// Action functions

static int doAction(Action::Result (Action::*toDo)(void) const, std::string failure_message) {
    if(!mavsdk_started) {
        log_error("Mavsdk not started");
        return -1;
    }

    const Action::Result result = (action->*toDo)();
    if(result != Action::Result::Success) {
        log_error_from_result(failure_message, result);
        return -1;
    }
    return 0;
}

static int doMavlinkCommand(MavlinkPassthrough::CommandLong command, std::string failure_message) {
    const MavlinkPassthrough::Result result = mavlink_passthrough->send_command_long(command);
    if(result != MavlinkPassthrough::Result::Success) {
        log_error_from_result(failure_message, result);
        return -1;
    }
    return 0;
}

// Coordinates related functions

static double toRad(double angle) {
    return M_PI * angle / 180;
}

static double toDeg(double angle) {
    return 180 * angle / M_PI;
}

/*
 * compute the bearing angle between point 1 and point 2
 * the bearing angle is the direction to take to go from point 1 to point 2
 * values are in [-π; π] where 0 is North
*/
static double bearing(double lat1, double lon1, double lat2, double lon2) {
    double lat1_rad = toRad(lat1);
    double lat2_rad = toRad(lat2);
    double dL = toRad(lon2 - lon1);
    double x = cos(lat2_rad) * sin(dL);
    double y = cos(lat1_rad) * sin(lat2_rad) - sin(lat1_rad) * cos(lat2_rad) * cos(dL);
    return atan2(x, y);
}

/*
 * To ensure that a drone goes through the target point and to avoid loitering
 * trajectory, the target point is projected to a distance equal to twice the
 * default loiter radius.
*/
static Coordinates getProjectionCoordinates(Coordinates destination, Coordinates position) {
    double laRad = toRad(destination.latitude);
    double sinLa = sin(laRad);
    double cosLa = cos(laRad);
    double loRad = toRad(destination.longitude);
    double bearing_angle = bearing(position.latitude, position.longitude,
                                   targeted_destination.latitude,
                                   targeted_destination.longitude);
    double newLa = asin(sinLa * COS_ADDED_DISTANCE
                        + cosLa * SIN_ADDED_DISTANCE * cos(bearing_angle));
    double newLo = loRad + atan2(sin(bearing_angle) * SIN_ADDED_DISTANCE * cosLa,
                                 COS_ADDED_DISTANCE - sinLa * sin(newLa));
    Coordinates projected{ toDeg(newLa), toDeg(newLo) };
    return projected;
}

static int doReposition(float la, float lo, float radius, float y) {
    if (!mavsdk_started)
        return -1;

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_REPOSITION;
    command.param1 = -1; // Ground speed, -1 for default
    command.param2 = MAV_DO_REPOSITION_FLAGS_CHANGE_MODE; //will go to navigate mode
    command.param3 = radius; // loiter radius
    command.param4 = y; // loiter direction, 0: clockwise 1: counter clockwise
    command.param5 = la;
    command.param6 = lo;
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();

    return doMavlinkCommand(command, "Entering loiter mode failed");
}

static void updateProjection(double current_lat, double current_lon) {
    Coordinates position{ current_lat, current_lon };
    projected_destination = getProjectionCoordinates(targeted_destination,
                                                     position);

    return doReposition(
        (float)projected_destination.latitude,
        (float)projected_destination.longitude,
        DEFAULT_RADIUS,
        0
    );
}

int updateLogAndProjection(void) {
    std::ostringstream oss;
    Telemetry::Position position;
    Telemetry::FixedwingMetrics metrics;

    if(mavsdk_started) {
        const auto now = std::chrono::system_clock::now();
        position = telemetry->position();
        metrics = telemetry->fixedwing_metrics();
        oss << std::chrono::duration_cast<std::chrono::milliseconds>(now.time_since_epoch()).count() << ";"
            << position.latitude_deg << ";" << position.longitude_deg << ";"
            << position.absolute_altitude_m << ";" << position.relative_altitude_m << ";"
            << telemetry->attitude_euler().yaw_deg << ";"
            << metrics.airspeed_m_s << ";" << metrics.climb_rate_m_s;
        log(oss.str());

        if (projected_destination.latitude != 0 || targeted_destination.latitude != 0) {
            return updateProjection(position.latitude_deg, position.longitude_deg);
        }
    }

    return 0;
}

// Connexion management functions

int start(const char * ip, int port, const char * log_file, int timeout) {
    std::string remote_ip(ip);
    ConnectionResult connection_result;
    float abs_altitude;
    log_file_fd.open(log_file);

    connection_result = _mavsdk.add_udp_connection(remote_ip, port);
    if (connection_result != ConnectionResult::Success) {
        log_error_from_result("Connection failed", connection_result);
        return -1;
    }

    log("Waiting to discover msystem...");
    fut = prom.get_future();

    _mavsdk.subscribe_on_new_system([]() {
        auto msystem_tmp = _mavsdk.systems().back();

        if (msystem_tmp->has_autopilot()) {
            log("Discovered autopilot");

            // Unsubscribe again as we only want to find one system.
            _mavsdk.subscribe_on_new_system(nullptr);
            prom.set_value(msystem_tmp);
        }
    });

    if (fut.wait_for(seconds(timeout)) == std::future_status::timeout) {
        log_error("No autopilot found, exiting.");
        return -1;
    }

    msystem = fut.get();
    telemetry = new Telemetry(msystem);
    action = new Action(msystem);
    mavlink_passthrough = new MavlinkPassthrough(msystem);

    do {
        log("Waiting for first telemetry");
        sleep_for(seconds(1));
        abs_altitude = getAltitude();
    } while(isnan(abs_altitude) || abs_altitude == 0);

    origin = telemetry->position();

    log("MAVSDK started...");
    mavsdk_started = 1;
    log("timestamp (ms);latitude (°);longitude (°);AMSL (m);rel altitude (m);yaw (°);air speed (m/s);climb rate (m/s)");

    return 0;
}

int stop() {
    // Delete pointers
    delete action;
    delete mavlink_passthrough;
    delete telemetry;

    log_file_fd << std::flush;
    log_file_fd.close();

    return 0;
}

int reboot() {
    return doAction(&Action::reboot, "Rebooting failed");
}

// Flight state management functions

int arm(void) {
    if(!mavsdk_started)
        return -1;

    while(!telemetry->health().is_home_position_ok) {
        log("Waiting for home position to be set");
        sleep_for(seconds(1));
    }

    log("Arming...");
    return doAction(&Action::arm, "Arming failed");
}

int takeOff(void) {
    if (doAction(&Action::takeoff, "Takeoff failed"))
        return -1;

    return 0;
}

int takeOffAndWait(void) {
    if (takeOff() < 0)
        return -1;

    while (telemetry->flight_mode() != Telemetry::FlightMode::Takeoff)
        sleep_for(seconds(1));
    while (telemetry->flight_mode() == Telemetry::FlightMode::Takeoff)
        sleep_for(seconds(1));

    return doAction(&Action::transition_to_fixedwing, "Transition to fixedwing failed");
}

int triggerParachute(void) {
    if (!mavsdk_started)
        return -1;

    log("Landing...");
    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_PARACHUTE;
    command.param1 = 2; //see https://mavlink.io/en/messages/common.html#PARACHUTE_ACTION
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();
    return doMavlinkCommand(command, "Parachute release failed");
}

// Flight management functions

static int setAltitude(float altitude) {
    if (!mavsdk_started)
        return -1;

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_OVERRIDE;
    command.param1 = 1 | 2 | 4 | 8;
    command.param2 = 1 | 2 | 8;
    command.param3 = altitude;
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();

    return doMavlinkCommand(command, "Setting altitude failed");
}

int loiter(double la, double lo, float a, float radius) {
    int result = doReposition((float)la, (float)lo, radius, 0);
    result |= setAltitude(a);
    return result;
}

int setAirspeed(float airspeed) {
    if (!mavsdk_started)
        return -1;

    MavlinkPassthrough::CommandLong command;
    command.command = MAV_CMD_DO_OVERRIDE;
    command.param1 = 1 | 2 | 4 | 8;
    command.param2 = 2 | 4 | 8;
    command.param4 = airspeed;
    command.target_sysid = mavlink_passthrough->get_target_sysid();
    command.target_compid = mavlink_passthrough->get_target_compid();

    return doMavlinkCommand(command, "Setting airspeed failed");
}

int setTargetCoordinates(double la, double lo, float a) {
    int result;
    if (!mavsdk_started)
        return -1;

    Telemetry::Position position = telemetry->position();
    targeted_destination.latitude = la;
    targeted_destination.longitude = lo;
    
    result = updateProjection(position.latitude_deg, position.longitude_deg);
    result |= setAltitude(a);
    return result;
}

// Information functions

float getAltitude(void) {
    return telemetry->position().absolute_altitude_m;
}

float getAltitudeRel(void) {
    return telemetry->position().relative_altitude_m;
}

float getInitialAltitude(void) {
    return origin.absolute_altitude_m;
}

double getInitialLatitude(void) {
    return origin.latitude_deg;
}

double getInitialLongitude(void) {
    return origin.longitude_deg;
}

double getLatitude(void) {
    return telemetry->position().latitude_deg;
}

double getLongitude(void) {
    return telemetry->position().longitude_deg;
}

double *getPositionArray(void) {
    Telemetry::Position position = telemetry->position();
    double *positionArray = (double*) malloc(POSITION_ARRAY_SIZE * sizeof(double));
    positionArray[0] = position.latitude_deg;
    positionArray[1] = position.longitude_deg;
    positionArray[2] = (double) position.absolute_altitude_m;
    positionArray[3] = (double) position.relative_altitude_m;
    return positionArray;
}

float *getSpeedArray(void) {
    Telemetry::FixedwingMetrics metrics = telemetry->fixedwing_metrics();
    float *speedArray = (float*) malloc(SPEED_ARRAY_SIZE * sizeof(float));
    speedArray[0] = getYaw();
    speedArray[1] = metrics.airspeed_m_s;
    speedArray[2] = metrics.climb_rate_m_s;
    return speedArray;
}

double getTakeOffAltitude(void) {
    const std::pair<Action::Result, float> response = action->get_takeoff_altitude();

    if (response.first != Action::Result::Success) {
        log_error_from_result("Get takeoff altitude failed", response.first);
        return -1;
    }
    return response.second;
}

float getYaw(void) {
    return telemetry->attitude_euler().yaw_deg;
}

float getSpeed(void) {
    return telemetry->fixedwing_metrics().airspeed_m_s;
}

float getClimbRate(void) {
    return telemetry->fixedwing_metrics().climb_rate_m_s;
}

int healthAllOk(void) {
    return telemetry->health_all_ok();
}
